import functools
import logging.config
import socket
from typing import Any
from typing import Dict
from typing import Tuple


__all__ = ["HostInformation", "DockerComposeInformation", "DockerInformation"]


@functools.lru_cache()
def _get_container_attributes(hostname: str) -> Dict[str, Any]:
    import docker

    try:
        client = docker.from_env()
        return client.containers.get(hostname).attrs
    except Exception:
        return dict()


class HostInformation(logging.Filter):
    """
    Add information about the flask app and configuration
    """

    def __init__(self) -> None:
        self.hostname = socket.gethostname()

    def filter(self, log_record: logging.LogRecord) -> bool:
        log_record.hostname = self.hostname  # type: ignore
        return True


class DockerInformation(logging.Filter):
    """
    Add information about the flask app and configuration
    """

    SCHEMA: Dict[Tuple[str, ...], Tuple[str, ...]] = {
        ("docker", "id"): ("Id",),
        ("docker", "name"): ("Name",),
        ("docker", "platform"): ("Platform",),
    }

    def __init__(self) -> None:

        docker_attributes = _get_container_attributes(socket.gethostname())

        filter_info: Dict[str, Any] = {}

        for target, source in self.SCHEMA.items():
            attrs = docker_attributes
            *parents, source_key = source
            for parent in parents:
                attrs = attrs.get(parent, {})

            if source_key not in attrs:
                continue

            value = attrs.get(source_key, None)

            attrs = filter_info
            *parents, target_key = target
            for parent in parents:
                attrs = filter_info.setdefault(parent, {})
            attrs[target_key] = value

        self.docker_attributes = filter_info

    def filter(self, log_record: logging.LogRecord) -> bool:
        for key, value in self.docker_attributes.items():
            setattr(log_record, key, value)
        return True


class DockerComposeInformation(DockerInformation):
    """
    Add information about the flask app and configuration
    """

    SCHEMA: Dict[Tuple[str, ...], Tuple[str, ...]] = {
        ("compose", "project"): ("Config", "Labels", "com.docker.compose.project"),
        ("compose", "service"): ("Config", "Labels", "com.docker.compose.service"),
        ("compose", "container-number"): ("Config", "Labels", "com.docker.compose.container-number"),
        ("compose", "config_files"): ("Config", "Labels", "com.docker.compose.project.config_files"),
    }



import logging.config
import time
import uuid
from typing import Any
from typing import Dict

from flask import current_app
from flask import Flask
from flask import g
from flask import has_app_context
from flask import has_request_context
from flask import request
from flask import Response

from .request_context import request_context_manger
from .request_context import RequestContextGenerator

__all__ = ["FlaskAppInformation", "RequestInformation", "log_request", "log_response"]


class FlaskAppInformation(logging.Filter):
    """
    Add information about the flask app and configuration
    """

    def filter(self, log_record: Any) -> bool:
        if has_app_context():
            log_record.flask = dict(
                environment=current_app.env, instance_path=current_app.instance_path, name=current_app.name
            )
        return True


class RequestInformation(logging.Filter):
    """
    Add request information to log records
    """

    def filter(self, log_record: Any) -> bool:
        """Enrich log records with request infomration"""
        if has_request_context():

            log_record.request = {
                "path": request.path,
                "method": request.method,
                "remote_addr": request.remote_addr,
                "user_agent": request.user_agent,
            }
            if "SERVER_PROTOCOL" in request.environ:
                log_record.request["protocol"] = request.environ["SERVER_PROTOCOL"]

            if hasattr(g, "request_id"):
                log_record.request["id"] = g.request_id

            log_record.url = request.path
            log_record.method = request.method
            log_record.remote_addr = request.remote_addr

        else:
            log_record.request = {}
            log_record.url = None
            log_record.method = None
            log_record.remote_addr = None

        return True


def log_request(sender: Flask, **extra: Any) -> None:
    """
    Log the start of a request to a flask app.
    """
    logger_name = sender.config.get("FLASK_LOGGING_REQUEST_LOGGER_NAME", "request")

    logger = sender.logger.getChild(logger_name)
    logger.debug(f"{request.method} {request.url} BEGIN", extra={"event": "request_started"})


def log_response(sender: Flask, response: Response, **extra: Any) -> None:
    """
    Log a response form the app.

    Should be attached to the `request_finished` signal.
    """
    logger_name = sender.config.get("FLASK_LOGGING_RESPONSE_LOGGER_NAME", "response")
    logger = sender.logger.getChild(logger_name)

    response_keywords: Dict[str, Any] = {}
    response_keywords["status"] = response.status
    response_keywords["status_code"] = response.status_code
    response_keywords["content_type"] = response.content_type

    content_length = response.content_length
    if content_length is not None:
        response_keywords["content_length"] = content_length

    if "_request_duration" in g:
        response_keywords["request_duration"] = g._request_duration

    logger.info(f"{response.status}", extra=dict(response=response_keywords, event="request_finished"))


@request_context_manger
def request_set_id() -> RequestContextGenerator:
    """
    Set a UUID in the header for each request.
    """

    request_header = current_app.config.get("FLASK_LOGGING_REQUEST_ID_HEADER", "X-Request-ID")

    request_id = request.headers.get(request_header, None)
    if request_id is None:
        request_id = str(uuid.uuid4())

    g.setdefault("request_id", request_id)
    response = yield
    response.headers.setdefault(request_header, g.get("request_id", request_id))
    return response


@request_context_manger
def request_track_time() -> RequestContextGenerator:
    """
    Track the duration of each request.
    """
    start_time = time.monotonic()
    response = yield
    duration = time.monotonic() - start_time
    g._request_duration = duration
    return response



import contextlib
import logging.config
from typing import Any
from typing import Dict
from typing import Iterator
from typing import Union

__all__ = ["MetadataFilter", "log_metadata"]


class MetadataFilter(logging.Filter):
    """
    A filter which imparts constant additional contextual information to log records as they are observed.

    Parameters
    ----------
    metadata: dict
        Dictionary of metadata to be added to all log records.

    """

    def __init__(self, metadata: Dict[str, Any]) -> None:
        self.metadata = metadata

    def filter(self, log_record: Any) -> bool:
        self._apply_metadata(log_record)
        return True

    def _apply_metadata(self, log_record: Any) -> None:
        log_record.__dict__.update(self.metadata)


@contextlib.contextmanager
def log_metadata(metadata: Dict[str, Any], logger: Union[None, str, logging.Filterer] = None) -> Iterator[None]:
    """
    Apply log metadata to all messages on the indicated handler or logger, for this context block

    Parameters
    ----------
    metadata: dict
        Dictionary of metadata to be added to all log records.
    logger: logger, handler, or logger name, optional.
        The logger to apply this filter to, defaults to the root logger.

    """

    if not isinstance(logger, logging.Filterer):
        logger = logging.getLogger(logger)

    metadata_filter = MetadataFilter(metadata)
    logger.addFilter(metadata_filter)

    yield

    logger.removeFilter(metadata_filter)



from .click import ClickStreamHandler
from .click import ClickStyleFormatter
from .click import StreamMode
from .json import JSONFormatter

__all__ = ["ClickStreamHandler", "ClickStyleFormatter", "StreamMode", "JSONFormatter"]



"""
Log iterator progress messages
"""
import dataclasses as dc
import logging
import time
from typing import Generic
from typing import Iterator
from typing import Optional
from typing import TypeVar
from typing import Union

__all__ = ["ProgressReporter"]

T = TypeVar("T")
LogLevel = Union[str, int]


@dc.dataclass
class ProgressReporter(Generic[T]):
    """
    Wraps an iterator which will report iteration progress to a logger.

    Parameters
    ----------
    iterator: iterator
        Target iterator, which will be consumed by this progress reporter.
    logger: logging.Logger compatible object
        Where to log progress messages
    total: int, optional
        Number of items in the iterator. If not provided, will call len(iterator)
        to find out. It is expected that unbounded iterators, or iterators which don't
        support len() are passed with a value for `total`.
    title: str, optional
        Message to include in logging messages, defaults to "items", but can be set to provide
        a more descriptive progress update.
    level: int or str, optional
        Logging level to use for progress messages
    progress_interval: float, optional
        How often (at best effort) to log iterator progress, in seconds. This wrapper does not use a thread, and so
        cannot pre-empt iteration. Default is 0.5 seconds.
    progress_every: int, optional
        How many elements to consume before logging progress. Setting a value here will log progress every N elements,
        irrespecitve of the time which has passed since the previous message. Defaults to `None` (i.e. relies on
        processing time above).
    progress_start: int, optional
        Minimum number of elements which triggers an initial log message. For short loops, printing both the start (0%)
        and end message can be unnecessary, so setting this parameter to `None`.

    """

    iterator: Iterator[T]
    logger: logging.Logger = dc.field(default_factory=logging.getLogger)
    total: int = -1
    title: str = "items"
    level: LogLevel = logging.INFO

    progress_interval: Optional[float] = 0.5
    progress_every: Optional[int] = None
    progress_start: Optional[int] = None

    def __post_init__(self) -> None:

        if self.total == -1:
            # Accept the type error here - this iterator wrapper isn't actually useful
            # if we can't know the total number of items we are iterating over.
            self.total = len(self.iterator)  # type: ignore

        self.total = max(self.total, 0)

    def __len__(self) -> int:
        return len(self.iterator)  # type: ignore

    @property
    def _loglevel(self) -> int:
        if isinstance(self.level, str):
            return getattr(logging, self.level.upper())
        return self.level

    def __iter__(self) -> Iterator[T]:

        message = start = time.monotonic()

        if self.progress_start is not None:
            if self.total and self.total >= self.progress_start:
                self.logger.log(self._loglevel, f"[  0%] Processed 0 of {self.total} {self.title}")

        for i, entity in enumerate(self.iterator, start=1):

            yield entity

            if (
                i
                and (self.progress_interval and (time.monotonic() - message) > self.progress_interval)
                or (self.progress_every and (i % self.progress_every == 0))
            ):
                message = time.monotonic()
                self.logger.log(
                    self._loglevel,
                    f"[{i/self.total: 4.0%}] Processed {i} of {self.total} {self.title}",
                    extra=dict(total=self.total, processed=i, duration=time.monotonic() - start),
                )

        duration = time.monotonic() - start
        self.logger.log(
            self._loglevel,
            f"[100%] Processed {self.total} of {self.total} {self.title} in {duration:.1f}s",
            extra=dict(total=self.total, duration=duration),
        )



import datetime as dt
import logging
from collections import defaultdict
from typing import DefaultDict


class CommonLogFormat(logging.Formatter):
    """Python standard logging formatter which matches the
    common log format which is the default for apache webservers."""

    def format(self, record: logging.LogRecord) -> str:
        timestamp = format_apache_clf_timestamp(record.created)
        request: DefaultDict[str, str] = defaultdict(lambda: "-", **getattr(record, "request", {}))
        response: DefaultDict[str, str] = defaultdict(lambda: "-", **getattr(record, "response", {}))

        return (
            f"{request['remote_addr']} {request['user_id']} {request['username']} [{timestamp}]"
            f' "{self.request_format(request)}"'
            f" {response['status_code']} {response['content_length']}"
        )

    def request_format(self, request: DefaultDict[str, str]) -> str:
        return f"{request['method']} {request['path']} {request['protocol']}"


def format_apache_clf_timestamp(timestamp: int) -> str:
    """Convert a timestamp (unix time, integer seconds) to the datetime format for CLF"""

    monthname = [None, "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

    when = dt.datetime.utcfromtimestamp(timestamp)

    s = "%02d/%3s/%04d %02d:%02d:%02d" % (
        when.day,
        monthname[when.month],
        when.year,
        when.hour,
        when.minute,
        when.second,
    )
    return s



import dataclasses as dc
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Tuple
from typing import Union
from weakref import WeakKeyDictionary

from flask import current_app
from flask import Flask
from flask import Response

__all__ = ["RequestContextWrapper", "request_context_manger"]

IntoResponse = Union[Response, Tuple[str, int], str]
RequestContextGenerator = Generator[Any, Response, Optional[IntoResponse]]
RequestContextFunction = Callable[[], RequestContextGenerator]
_ContextWrappers = Dict["RequestContextWrapper", RequestContextGenerator]

EXTENSION_NAME = "flask_logging.request_context"


@dc.dataclass
class _RequestContextState:
    context_wrappers: _ContextWrappers = dc.field(default_factory=lambda: cast(_ContextWrappers, WeakKeyDictionary()))


class RequestContextWrapper:
    """
    Decorator to set up a single generator function as a request wrapper.

    The generator will be sent the response at the end of the request, so it should yield
    at the point it would like the request to process.

    A minimal use::

        @RequestContextWrapper
        def request_track_time() -> RequestContextGenerator:
            start_time = time.monotonic()
            response = yield
            duration = time.monotonic() - start_time
            response.headers['X-Request-Duration'] = duration
            return response

    Parameters
    ----------
    context_function: callable generator
        Generator used for the request context.

    """

    def __init__(self, context_function: RequestContextFunction, app: Optional[Flask] = None) -> None:

        self.context = context_function
        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        app.before_request(self._before_request)
        app.after_request(self._after_request)
        app.extensions[EXTENSION_NAME] = _RequestContextState()

    def _get_state(self) -> _RequestContextState:
        try:
            if current_app:
                return current_app._get_current_object().extensions[EXTENSION_NAME]

            if self.app is not None:
                return self.app.extensions[EXTENSION_NAME]
        except KeyError:
            raise RuntimeError(
                "The request context extension was not registered to the current "
                "application.  Please make sure to call init_app() first."
            )

        raise RuntimeError("No application found. Either work inside a view function or push an application context")

    def _before_request(self) -> None:

        ctx = self.context()
        try:
            next(ctx)
        except StopIteration as e:
            raise RuntimeError("Generator did not yield") from e
        else:
            state = self._get_state()
            state.context_wrappers[self] = ctx

    def _after_request(self, response: Response) -> Response:
        state = self._get_state()

        try:
            ctx = state.context_wrappers.pop(self)
        except KeyError:
            # Can't find this context, so we don't modify the response and just return it here.
            return response

        try:
            ctx.send(response)
        except StopIteration as e:
            if isinstance(e.value, Response):
                response = e.value
            elif e.value is not None:
                raise RuntimeError(f"Generator returned something which is not a response: {type(e.value)}")
        else:
            raise RuntimeError("Generator did not stop")
        return response


def request_context_manger(f: RequestContextFunction) -> RequestContextWrapper:
    """
    Mark this function as a request context manager
    """
    return RequestContextWrapper(f, app=None)



import functools
import logging.config
import os
from pathlib import Path
from typing import Dict
from typing import Union

import yaml


__all__ = ["setup_null_handler", "configure_logging"]


def setup_null_handler(logger: str) -> None:
    """Add a null handler to a given logger"""
    log = logging.getLogger(logger)
    log.propagate = True
    log.addHandler(logging.NullHandler())


@functools.singledispatch
def configure_logging(cfg: Dict) -> None:
    """Configure logging via a configuration dictionary"""
    cfg.setdefault("disable_existing_loggers", False)
    logging.config.dictConfig(cfg)


@configure_logging.register(os.PathLike)
@configure_logging.register(str)
def _cl_filename(filename: "Union[os.PathLike, str]") -> None:
    """Configure logging via a configuration in a YAML file"""
    with Path(filename).open("r") as s:
        cfg = yaml.safe_load(s)

    configure_logging(cfg)



import logging
import os
from typing import Optional

import pkg_resources
from flask import Config
from flask import Flask
from flask.signals import request_finished
from flask.signals import request_started

from .config import configure_logging
from .flask import FlaskAppInformation
from .flask import log_request
from .flask import log_response
from .flask import request_set_id
from .flask import request_track_time
from .flask import RequestInformation


class FlaskLogging:
    """
    Core Flask-Logging extension object.

    This class wraps several features of this extension in a re-usable, configurable package
    """

    def __init__(self, app: Optional[Flask] = None) -> None:
        self.app = app
        if app is not None:
            self.init_app(app)

    def add_flask_filters(self, logger: logging.Logger) -> None:
        if not any(isinstance(f, RequestInformation) for f in logger.filters):
            logger.addFilter(RequestInformation())
        if not any(isinstance(f, FlaskAppInformation) for f in logger.filters):
            logger.addFilter(FlaskAppInformation())

    def set_defaults(self, app: Flask) -> None:
        default_filename = os.path.abspath(pkg_resources.resource_filename(__name__, "defaults.cfg"))
        defaults = Config(root_path="/dev/null")
        defaults.from_pyfile(default_filename)
        for key, value in defaults.items():
            app.config.setdefault(key, value)

    def init_app(self, app: Flask) -> None:
        self.set_defaults(app)

        if app.config.get("FLASK_LOGGING_CONFIGURATION"):
            configure_logging(app.config["FLASK_LOGGING_CONFIGURATION"])

        if app.config.get("FLASK_LOGGING_REQUEST_ID"):
            request_set_id.init_app(app)

        if app.config.get("FLASK_LOGGING_REQUEST_DURATION"):
            request_track_time.init_app(app)

        if app.config.get("FLASK_LOGGING_REQUEST_FINISHED"):
            request_finished.connect(log_response, app)
        if app.config.get("FLASK_LOGGING_REQUEST_STARTED"):
            request_started.connect(log_request, app)

        if app.config.get("FLASK_LOGGING_REQUEST_LOGGER"):
            logger_name = app.config.get("FLASK_LOGGING_REQUEST_LOGGER_NAME", "request")
            self.add_flask_filters(app.logger.getChild(logger_name))

        if app.config.get("FLASK_LOGGING_RESPONSE_LOGGER"):
            logger_name = app.config.get("FLASK_LOGGING_RESPONSE_LOGGER_NAME", "response")
            self.add_flask_filters(app.logger.getChild(logger_name))



import logging
from typing import Any
from typing import Callable
from typing import Dict
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union
from urllib.parse import parse_qs
from urllib.parse import urlencode
from urllib.parse import urlparse
from urllib.parse import urlunparse

if TYPE_CHECKING:
    import redis

from .json import makeLogRecordfromJson

ClientArgs = Union["redis.Redis", "redis.ConnectionPool", Tuple[str, int], str]
Deserialize = Callable[[str], logging.LogRecord]

DESERIALIERS: Dict[str, Deserialize] = {"json": makeLogRecordfromJson}


def _handle_redis_client_args(args: ClientArgs) -> "redis.Redis":
    """Handle arguments that should produce a REDIS client."""
    import redis

    if isinstance(args, redis.Redis):
        client: "redis.Redis" = args
    elif isinstance(args, redis.ConnectionPool):
        client = redis.Redis(connection_pool=args, decode_responses=False)
    elif isinstance(args, tuple):
        host, port = args
        client = redis.Redis(host, port, decode_responses=False)
    elif isinstance(args, str):
        client = redis.Redis.from_url(args)
    else:
        raise TypeError(f"Can't handle client arguments: {args!r}")
    return client


class RedisPublisher(logging.Handler):
    """A Redis publisher, which takes formatted log messages and publishes them to Redis."""

    def __init__(self, address: ClientArgs, channel: str):
        super().__init__()
        self.client = _handle_redis_client_args(address)
        self.channel = channel

    def emit(self, record: logging.LogRecord) -> None:
        """Emit a single record."""
        try:
            msg = self.format(record)
            self.client.publish(self.channel, msg)
            self.flush()
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException:
            self.handleError(record)


class RedisLogWatcher:
    """Watch a Redis channel for logging"""

    thread = None

    def __init__(
        self,
        address: ClientArgs,
        channel: str,
        deserialize: Deserialize,
    ) -> None:
        super().__init__()
        self.pubsub = _handle_redis_client_args(address).pubsub()
        self.subscribe(channel)
        self.deserialize = deserialize

    @classmethod
    def from_url(cls, url: str) -> "RedisLogWatcher":
        """Create the log watcher from a URL"""

        # Get the options out of the URL
        result = urlparse(url)
        options = parse_qs(result.query)
        channel = options.pop("channel", "")
        format = options.pop("format", "json")
        if isinstance(format, list):
            raise TypeError(f"Multiple formats provided in URL: {format!r}")

        deserialize = DESERIALIERS[format]

        # Rebuild the URL without the query string.
        args = list(result)
        if "db" in options:
            args[4] = urlencode(options)
        else:
            args[4] = ""
        url = urlunparse(args)

        # Set up the object.
        if isinstance(channel, list):
            channel, *rest = channel
            obj = cls(url, channel, deserialize)
            for channel in rest:
                obj.subscribe(channel)
        else:
            obj = cls(url, channel, deserialize)
        return obj

    def _redis_responder(self, msg):
        """Given a Redis message, create the logrecord and handle it."""
        record = self.deserialize(msg["data"])
        logging.getLogger(record.name).handle(record)

    def subscribe(self, name):
        """Subscribe to an addtional channel."""
        self.pubsub.subscribe(**{name: self._redis_responder})

    def start(self):
        """Start the log watcher."""
        self.thread = self.pubsub.run_in_thread(sleep_time=0.01)

    def stop(self):
        """Stop the log watcher"""
        if self.thread is not None:
            self.thread.stop()

    def __enter__(self) -> "RedisLogWatcher":
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()



import logging
import threading
from typing import Any
from typing import Iterator
from typing import List
from typing import MutableMapping
from typing import Tuple
from typing import TypeVar
from typing import Union

LogLevel = Union[str, int]
V = TypeVar("V")

__all__ = ["LogLevelDict"]


class LogLevelDict(MutableMapping[LogLevel, V]):
    """
    A dictionary for looking up logging properties based on log level

    Looking up an intermediate level will return the record for the next
    highest level. If the dictionary only contains values for INFO (20)
    and CRITICAL (50), looking up WARNING will return the value for
    CRITICAL, and looking up DEBUG or INFO will return the value for INFO.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()
        self._lock = threading.Lock()
        self._items: List[Tuple[int, V]] = sorted((self._corece_key(k), v) for k, v in dict(*args, **kwargs).items())

    def _corece_key(self, key: LogLevel) -> int:
        if isinstance(key, str):
            key = getattr(logging, key.upper())
        return int(key)

    def __getitem__(self, key: LogLevel) -> V:
        key = self._corece_key(key)
        with self._lock:
            for lvl, value in self._items:
                if lvl >= key:
                    return value
        raise KeyError(key)

    def __setitem__(self, key: LogLevel, value: V) -> None:
        key = self._corece_key(key)

        replace = False
        with self._lock:
            idx = 0
            for idx, (lvl, _) in enumerate(self._items):
                if lvl == key:
                    replace = True
                    break
                elif lvl > key:
                    break
            if replace:
                del self._items[idx]
            self._items.insert(idx, (key, value))

    def __delitem__(self, key: LogLevel) -> None:
        key = self._corece_key(key)

        with self._lock:
            idx = 0
            for idx, (lvl, _) in enumerate(self._items):
                if lvl == key:
                    break
            else:
                raise KeyError(key)
            del self._items[idx]

    def __iter__(self) -> Iterator[int]:
        for lvl, _ in self._items:
            yield lvl

    def __len__(self) -> int:
        return len(self._items)

    def __repr__(self) -> str:
        return "LogLevelDict({!r})".format(dict(self._items))



import abc
import datetime as dt
import enum
import json
import logging.config
import uuid
import warnings
from typing import Any
from typing import Callable
from typing import Dict
from typing import Tuple
from typing import Type
from typing import Union

from werkzeug.local import LocalProxy
from werkzeug.useragents import UserAgent


__all__ = ["JSONLogWarning", "JSONFormatter"]


LOG_RECORD_SCHEMA: Dict[str, Tuple[str, ...]] = {
    "message": ("message",),
    "msg": ("message_info", "text"),
    "args": ("message_info", "args"),
    "asctime": ("timing", "ascii"),
    "created": ("timing", "created"),
    "msecs": ("timing", "msecs"),
    "relativeCreated": ("timing", "relativeCreated"),
    "pathname": ("python", "code", "pathname"),
    "filename": ("python", "code", "filename"),
    "module": ("python", "code", "module"),
    "lineno": ("python", "code", "lineno"),
    "funcName": ("python", "code", "funcname"),
    "stack_info": ("python", "stack"),
    "exc_info": ("python", "exc", "info"),
    "exc_text": ("python", "exc", "text"),
    "threadName": ("python", "thread", "name"),
    "thread": ("python", "thread", "id"),
    "process": ("python", "process", "pid"),
    "processName": ("python", "process", "name"),
    "name": ("logger", "name"),
    "levelno": ("logger", "level", "number"),
    "levelname": ("logger", "level", "name"),
    "clevelname": ("logger", "level", "ansiname"),
}


class JSONLogWarning(Warning):
    """Warning used when an unmarshallable type is being logged"""


class HasSchema(abc.ABC):
    @classmethod
    def __subclasshook__(cls: Type["HasSchema"], C: Type) -> bool:
        if cls is HasSchema:
            if any("__schema__" in B.__dict__ for B in C.__mro__):
                return True
        return NotImplemented


class FlexJSONEncoder(json.JSONEncoder):
    """Flexible JSON encoder for use with more JSON types"""

    converters: Dict[Union[Type, Tuple[Type]], Callable[[Any], Any]] = {
        dt.datetime: lambda value: value.isoformat(),
        dt.date: lambda value: f"{value:%Y-%m-%d}",
        HasSchema: lambda m: m.__schema__().dump(m),
        uuid.UUID: str,
        UserAgent: str,
        enum.Enum: lambda value: value.name,
        LocalProxy: repr,
    }

    def default(self, obj: Any) -> Any:
        for clses, func in self.converters.items():
            if isinstance(obj, clses):
                return func(obj)
        else:
            warnings.warn(JSONLogWarning(f"Unable to marshal type {type(obj)} to JSON"))
            return f"<Unecodeable type: {type(obj)!r} {obj!r}>"
        return super().default(obj)


class JSONFormatter(logging.Formatter):
    """
    Format log records as JSON
    """

    def format(self, record: logging.LogRecord) -> str:
        """
        Format a record for output
        """
        ei = record.exc_info
        if ei:
            _ = super().format(record)  # just to get traceback text into record.exc_text
            record.exc_info = None  # to avoid Unpickleable error
        s = json.dumps(self._convert_json_data(record), sort_keys=True, cls=FlexJSONEncoder)
        if ei:
            record.exc_info = ei  # for next handler
        return s

    def _convert_json_data(self, record: logging.LogRecord) -> Dict[str, Any]:
        data: Dict[str, Any] = {}

        raw = dict(record.__dict__)

        for key, value in raw.items():
            *parents, target_key = LOG_RECORD_SCHEMA.get(key, (key,))
            target = data
            for parent in parents:
                target = target.setdefault(parent, {})
            target[target_key] = value

        return data


def makeLogRecordfromJson(data: str) -> logging.LogRecord:
    raw = json.loads(data)
    recordinfo = {**raw}

    if isinstance(raw.get("message", None), dict):
        message = raw.pop("message")
        if "message_info" not in raw:
            raw["message_info"] = message
        else:
            raw["message_original"] = message

    # This will duplicate standard keys – i.e. it doesn't
    # remove the nested ones, but thats probably fine?
    for key, position in LOG_RECORD_SCHEMA.items():

        *parents, target_key = position
        target = raw
        for parent in parents:
            target = target.get(parent, {})

        if target_key in target:
            recordinfo[key] = target[target_key]

    if "args" in recordinfo:
        recordinfo["args"] = tuple(recordinfo["args"])

    if "message" in recordinfo and "msg" not in recordinfo:
        recordinfo["msg"] = recordinfo["message"]
        recordinfo["args"] = None

    return logging.makeLogRecord(recordinfo)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['flask_logging', 'flask_logging.handlers']

package_data = \
{'': ['*']}

install_requires = \
['blinker>=1.4,<2.0', 'flask>=1.1.2,<2.0.0', 'pyyaml>=5.3.1,<6.0.0']

extras_require = \
{'docker': ['docker>=4.4.4,<5.0.0'], 'redis': ['redis>=3.5.3,<4.0.0']}

setup_kwargs = {
    'name': 'flask-logging-helpers',
    'version': '0.4.2',
    'description': 'Logging tools for flask',
    'long_description': None,
    'author': 'Alex Rudy',
    'author_email': 'opensource@alexrudy.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import enum
import itertools
import logging.config
from typing import Any
from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import Union

import click

from ..loglevel import LogLevelDict

__all__ = ["ClickStyleFormatter", "ClickStreamHandler", "StreamMode"]

LogLevel = Union[str, int]

LEVEL_STYLES: LogLevelDict[Dict[str, Any]] = LogLevelDict(
    {
        logging.DEBUG: dict(fg="blue"),
        logging.INFO: dict(fg="green"),
        logging.WARNING: dict(fg="yellow"),
        logging.ERROR: dict(fg="red"),
        logging.CRITICAL: dict(fg="red"),
    }
)

COLORS = ["green", "yellow", "blue", "magenta", "cyan", "red"]


class ClickStyleFormatter(logging.Formatter):
    """
    Format log record attributes based on click style parameters.
    """

    def __init__(
        self,
        fmt: Optional[str] = None,
        datefmt: Optional[str] = None,
        style: str = "%",
        level_styles: Optional[Dict[LogLevel, Dict[str, Any]]] = None,
        attributes: Optional[List[str]] = None,
    ) -> None:
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)
        self.level_styles = LogLevelDict(level_styles or LEVEL_STYLES)  # type: ignore
        self.attributes = attributes or ["levelname"]

    def apply_style_attributes(self, record: logging.LogRecord) -> logging.LogRecord:
        """
        Apply style attributes based on a log level.
        """
        for attribute in self.attributes:
            value = getattr(record, attribute)
            color_value = self.style_string(record.levelno, value)
            setattr(record, f"c{attribute}", color_value)
        return record

    def format(self, record: logging.LogRecord) -> str:
        """
        Format a record for output
        """
        record = self.apply_style_attributes(record)
        return super().format(record)

    def style_string(self, level: LogLevel, message: str) -> str:
        """
        Style a message with the appropriate settings
        """
        style = self.level_styles.get(level, {})
        return click.style(message, **style)


class StreamMode(enum.Enum):
    STDOUT = "stdout"
    STDERR = "stderr"
    SPLIT = "split"


class ClickStreamHandler(logging.StreamHandler):
    def __init__(
        self, level: LogLevel = logging.NOTSET, mode: StreamMode = StreamMode.SPLIT, split: LogLevel = logging.INFO
    ) -> None:
        super().__init__()
        self.setLevel(level)

        stdout = click.get_text_stream("stdout")
        stderr = click.get_text_stream("stderr")

        streams: LogLevelDict[IO[str]]
        if mode == StreamMode.SPLIT:
            streams = LogLevelDict({split: stdout, logging.CRITICAL: stderr})
        elif mode == StreamMode.STDOUT:
            streams = LogLevelDict({logging.CRITICAL: stdout})
        elif mode == StreamMode.STDERR:
            streams = LogLevelDict({logging.CRITICAL: stderr})
        self.streams = streams

    def flush(self) -> None:
        """
        Flushes the stream.
        """
        self.acquire()
        try:
            for stream in self.streams.values():
                if hasattr(stream, "flush"):
                    stream.flush()
        finally:
            self.release()

    def emit(self, record: logging.LogRecord) -> None:
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        The record is then written to the stream with a trailing newline.  If
        exception information is present, it is formatted using
        traceback.print_exception and appended to the stream.  If the stream
        has an 'encoding' attribute, it is used to determine how to do the
        output to the stream.
        """
        try:
            stream = self.streams[record.levelno]
            msg = self.format(record)
            stream.write(msg + self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)

    @classmethod
    def from_format_and_settings(
        cls,
        level: LogLevel = logging.NOTSET,
        mode: StreamMode = StreamMode.SPLIT,
        split: LogLevel = logging.INFO,
        fmt: Optional[str] = None,
        datefmt: Optional[str] = None,
        style: str = "%",
        level_styles: Optional[Dict[LogLevel, Dict[str, Any]]] = None,
        attributes: Optional[List[str]] = None,
    ) -> "ClickStreamHandler":
        """
        Set up a click stream handler and a click color formatter for logging.
        """
        formatter = ClickStyleFormatter(
            fmt=fmt, datefmt=datefmt, style=style, level_styles=level_styles, attributes=attributes
        )
        handler = cls(level=level, mode=mode, split=split)
        handler.setFormatter(formatter)
        return handler


class ComposeClickFormatter(logging.Formatter):
    def __init__(self, datefmt: Optional[str] = None) -> None:
        super().__init__(fmt=None, datefmt=datefmt)

        self._colors = itertools.cycle(COLORS)
        self._service_colors: Dict[str, str] = {}

    def format(self, record: logging.LogRecord) -> str:
        _ = super().format(record)

        service = getattr(record, "compose", {}).get("service", "")
        number = int(getattr(record, "compose", {}).get("container-number", 0))

        prefix_parts = []

        level_style = LEVEL_STYLES[record.levelno]
        c_level = click.style(f"{record.levelname:^11.11s}", **level_style)

        if service:
            if service not in self._service_colors:
                self._service_colors[service] = next(self._colors)
            service_color = self._service_colors[service]
            c_service = click.style(service, fg=service_color)
            prefix_parts.append(c_service)

        if number:
            prefix_parts.append(f"{number:d}")

        if len(record.name) > 20:
            name = "..." + record.name[-17:]
        else:
            name = f"{record.name:20.20s}"
        prefix_parts.append(name)
        prefix_parts.append(c_level)

        prefix = f"[{'|'.join(prefix_parts)}]"

        record.asctime = self.formatTime(record, self.datefmt)
        formatted_message = self.formatMessage(record)

        if not formatted_message:
            print(record.__dict__)

        record.message = formatted_message

        message = "{prefix} {message:s} [{asctime:s}]".format_map({**record.__dict__, "prefix": prefix})

        return message



from .config import configure_logging
from .config import setup_null_handler
from .core import FlaskLogging
from .filters import log_metadata
from .flask import FlaskAppInformation
from .flask import RequestInformation
from .handlers import ClickStreamHandler
from .handlers import ClickStyleFormatter
from .handlers import JSONFormatter
from .loglevel import LogLevelDict

__all__ = [
    "LogLevelDict",
    "configure_logging",
    "ClickStreamHandler",
    "ClickStyleFormatter",
    "JSONFormatter",
    "log_metadata",
    "setup_null_handler",
    "FlaskAppInformation",
    "RequestInformation",
    "FlaskLogging",
]
